@@ -2,7 +2,7 @@
  * Unsquash a squashfs filesystem.  This is a highly compressed read only
  * filesystem.
  *
- * Copyright (c) 2009, 2010, 2013, 2019, 2021
+ * Copyright (c) 2009, 2010, 2012, 2013, 2019, 2021, 2022, 2023, 2025
  * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
@@ -25,10 +25,12 @@
 #include "unsquashfs.h"
 #include "squashfs_compat.h"
 #include "compressor.h"
+#include "alloc.h"
 
 static squashfs_fragment_entry_2 *fragment_table;
 static unsigned int *uid_table, *guid_table;
 static squashfs_operations ops;
+static int needs_sorting = FALSE;
 
 
 static void read_block_list(unsigned int *block_list, long long start,
@@ -39,9 +41,7 @@ static void read_block_list(unsigned int *block_list, long long start,
 	TRACE("read_block_list: blocks %d\n", blocks);
 
 	if(swap) {
-		char *block_ptr = malloc(blocks * sizeof(unsigned int));
-		if(block_ptr == NULL)
-			MEM_ERROR();
+		char *block_ptr = MALLOC(blocks * sizeof(unsigned int));
 		res = read_inode_data(block_ptr, &start, &offset, blocks * sizeof(unsigned int));
 		if(res == FALSE)
 			EXIT_UNSQUASH("read_block_list: failed to read "
@@ -85,19 +85,11 @@ static int read_fragment_table(long long *table_start)
 		"from 0x%llx\n", sBlk.s.fragments, indexes,
 		sBlk.s.fragment_table_start);
 
-	fragment_table_index = malloc(length);
-	if(fragment_table_index  == NULL)
-		MEM_ERROR();
-
-	fragment_table = malloc(bytes);
-	if(fragment_table == NULL)
-		MEM_ERROR();
+	fragment_table_index = MALLOC(length);
+	fragment_table = MALLOC(bytes);
 
 	if(swap) {
-		 unsigned int *sfragment_table_index = malloc(length);
-
-		if(sfragment_table_index == NULL)
-			MEM_ERROR();
+		 unsigned int *sfragment_table_index = MALLOC(length);
 
 		 res = read_fs_bytes(fd, sBlk.s.fragment_table_start,
 			length, sfragment_table_index);
@@ -190,12 +182,24 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 		EXIT_UNSQUASH("read_inode: failed to read inode %lld:%d\n", st, off);
 
 	i.xattr = SQUASHFS_INVALID_XATTR;
+
+	if(header.base.uid >= sBlk.no_uids)
+		EXIT_UNSQUASH("File system corrupted - uid index in inode too large (uid: %u)\n", header.base.uid);
+
 	i.uid = (uid_t) uid_table[header.base.uid];
-	i.gid = header.base.guid == SQUASHFS_GUIDS ? i.uid :
-		(uid_t) guid_table[header.base.guid];
+
+	if(header.base.guid == SQUASHFS_GUIDS)
+		i.gid = i.uid;
+	else if(header.base.guid >= sBlk.no_guids)
+		EXIT_UNSQUASH("File system corrupted - gid index in inode too large (gid: %d)\n", header.base.guid);
+	else
+		i.gid = (uid_t) guid_table[header.base.guid];
+
+	if(header.base.inode_type < 1 || header.base.inode_type > 8)
+		EXIT_UNSQUASH("File system corrupted - invalid type in inode (type: %u)\n", header.base.inode_type);
+
 	i.mode = lookup_type[header.base.inode_type] | header.base.mode;
 	i.type = header.base.inode_type;
-	i.time = sBlk.s.mkfs_time;
 	i.inode_number = inode_number++;
 
 	switch(header.base.inode_type) {
@@ -218,7 +222,10 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 			i.data = inode->file_size;
 			i.offset = inode->offset;
 			i.start = inode->start_block;
-			i.time = inode->mtime;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = inode->mtime;
 			break;
 		}
 		case SQUASHFS_LDIR_TYPE: {
@@ -240,7 +247,10 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 			i.data = inode->file_size;
 			i.offset = inode->offset;
 			i.start = inode->start_block;
-			i.time = inode->mtime;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = inode->mtime;
 			break;
 		}
 		case SQUASHFS_FILE_TYPE: {
@@ -260,7 +270,10 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 					"inode %lld:%d\n", start, offset);
 
 			i.data = inode->file_size;
-			i.time = inode->mtime;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = inode->mtime;
 			i.frag_bytes = inode->fragment == SQUASHFS_INVALID_FRAG
 				?  0 : inode->file_size % sBlk.s.block_size;
 			i.fragment = inode->fragment;
@@ -292,16 +305,17 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 				EXIT_UNSQUASH("read_inode: failed to read "
 					"inode %lld:%d\n", start, offset);
 
-			i.symlink = malloc(inodep->symlink_size + 1);
-			if(i.symlink == NULL)
-				MEM_ERROR();
-
+			i.symlink = MALLOC(inodep->symlink_size + 1);
 			res = read_inode_data(i.symlink, &start, &offset, inodep->symlink_size);
 			if(res == FALSE)
 				EXIT_UNSQUASH("read_inode: failed to read "
 					"inode symbolic link %lld:%d\n", start, offset);
 			i.symlink[inodep->symlink_size] = '\0';
 			i.data = inodep->symlink_size;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = sBlk.s.mkfs_time;
 			break;
 		}
  		case SQUASHFS_BLKDEV_TYPE:
@@ -322,11 +336,19 @@ static struct inode *read_inode(unsigned int start_block, unsigned int offset)
 					"inode %lld:%d\n", start, offset);
 
 			i.data = inodep->rdev;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = sBlk.s.mkfs_time;
 			break;
 			}
 		case SQUASHFS_FIFO_TYPE:
 		case SQUASHFS_SOCKET_TYPE:
 			i.data = 0;
+			if(time_opt)
+				i.time = timeval;
+			else
+				i.time = sBlk.s.mkfs_time;
 			break;
 		default:
 			EXIT_UNSQUASH("Unknown inode type %d in "
@@ -347,7 +369,7 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 	long long start;
 	int bytes = 0;
 	int dir_count, size, res;
-	struct dir_ent *new_dir;
+	struct dir_ent *ent, *cur_ent = NULL;
 	struct dir *dir;
 
 	TRACE("squashfs_opendir: inode start block %d, offset %d\n",
@@ -355,12 +377,9 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 
 	*i = read_inode(block_start, offset);
 
-	dir = malloc(sizeof(struct dir));
-	if(dir == NULL)
-		MEM_ERROR();
-
+	dir = MALLOC(sizeof(struct dir));
 	dir->dir_count = 0;
-	dir->cur_entry = 0;
+	dir->cur_entry = NULL;
 	dir->mode = (*i)->mode;
 	dir->uid = (*i)->uid;
 	dir->guid = (*i)->gid;
@@ -444,29 +463,37 @@ static struct dir *squashfs_opendir(unsigned int block_start, unsigned int offse
 				"%d:%d, type %d\n", dire->name,
 				dirh.start_block, dire->offset, dire->type);
 
-			if((dir->dir_count % DIR_ENT_SIZE) == 0) {
-				new_dir = realloc(dir->dirs, (dir->dir_count +
-					DIR_ENT_SIZE) * sizeof(struct dir_ent));
-				if(new_dir == NULL)
-					MEM_ERROR();
-				dir->dirs = new_dir;
-			}
-
-			strcpy(dir->dirs[dir->dir_count].name, dire->name);
-			dir->dirs[dir->dir_count].start_block =
-				dirh.start_block;
-			dir->dirs[dir->dir_count].offset = dire->offset;
-			dir->dirs[dir->dir_count].type = dire->type;
+			ent = MALLOC(sizeof(struct dir_ent));
+			ent->name = STRDUP(dire->name);
+			ent->start_block = dirh.start_block;
+			ent->offset = dire->offset;
+			ent->type = dire->type;
+			ent->next = NULL;
+			if(cur_ent == NULL)
+				dir->dirs = ent;
+			else
+				cur_ent->next = ent;
+			cur_ent = ent;
 			dir->dir_count ++;
 			bytes += dire->size + 1;
 		}
 	}
 
+	if(needs_sorting)
+		sort_directory(&(dir->dirs), dir->dir_count);
+
+	/* check directory for duplicate names and sorting */
+	if(check_directory(dir) == FALSE) {
+		if(needs_sorting)
+			ERROR("File system corrupted: directory has duplicate names\n");
+		else
+			ERROR("File system corrupted: directory has duplicate names or is unsorted\n");
+		goto corrupted;
+	}
 	return dir;
 
 corrupted:
-	free(dir->dirs);
-	free(dir);
+	squashfs_closedir(dir);
 	return NULL;
 }
 
@@ -596,6 +623,10 @@ int read_super_2(squashfs_operations **s_ops, void *s)
 	 * 2.x filesystems use gzip compression.
 	 */
 	comp = lookup_compressor("gzip");
+
+	if(sBlk_3->s_minor == 0)
+		needs_sorting = TRUE;
+
 	return TRUE;
 }
 
